package movies

import (
	"fmt"
	"os"
	"path/filepath"

	"github.com/Sirupsen/logrus"
	"github.com/jmoiron/sqlx"
	"github.com/odwrtw/polochon/lib"
	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/sqly"
	"gitlab.quimbo.fr/odwrtw/canape-sql/src/internal/web"
)

const (
	addMovieQuery = `
	INSERT INTO movies (imdb_id, title, rating, votes, plot, tmdb_id, year, original_title, runtime, sort_title, tagline)
	VALUES (:imdbid, :title, :rating, :votes, :plot, :tmdbid, :year, :originaltitle, :runtime, :sorttitle, :tagline)
	RETURNING id;`

	updateMovieQuery = `
	UPDATE movies
	SET imdb_id=:imdbid, title=:title, rating=:rating, votes=:votes, plot=:plot, tmdb_id=:tmdbid, year=:year, original_title=:originaltitle, runtime=:runtime, sort_title=:sorttitle, tagline=:tagline
	WHERE ID = :id;`

	getMovieQueryByImdbID = `
	SELECT
	id, imdb_id AS imdbid, title, rating, votes, plot,
	tmdb_id AS tmdbid, year,
	original_title AS originaltitle, runtime,
	sort_title AS sorttitle, tagline,
	created_at, updated_at
	FROM movies WHERE imdb_id=$1;`
	getMovieQueryByID = `
	SELECT
	id, imdb_id AS imdbid, title, rating, votes, plot,
	tmdb_id AS tmdbid, year,
	original_title AS originaltitle, runtime,
	sort_title AS sorttitle, tagline,
	created_at, updated_at
	FROM movies WHERE id=$1;`
	deleteMovieQuery = `DELETE FROM movies WHERE id=$1;`
)

var (
	// ErrNotFound error returned when show not found in database
	ErrNotFound = fmt.Errorf("Not found")
)

// Movie represents a movie
type Movie struct {
	sqly.BaseModel
	polochon.Movie
	PolochonURL string `json:"polochon_url"`
	PosterURL   string `json:"poster_url"`
}

func New(imdbID string) *Movie {
	return &Movie{
		Movie: polochon.Movie{
			ImdbID: imdbID,
		},
	}
}

// Get returns show details in database from id or imdbid or an error
func (m *Movie) Get(env *web.Env) error {
	var err error
	if m.ID != "" {
		err = env.Database.QueryRowx(getMovieQueryByID, m.ID).StructScan(m)
	} else if m.ImdbID != "" {
		err = env.Database.QueryRowx(getMovieQueryByImdbID, m.ImdbID).StructScan(m)
	} else {
		err = fmt.Errorf("Can't get movie details, you have to specify an ID or ImdbID")
	}
	if err != nil {
		if err.Error() == "sql: no rows in result set" {
			return ErrNotFound
		}
		return err
	}

	// Set the poster url
	m.PosterURL = m.GetPosterURL(env)

	return nil
}

// GetDetails retrieves details for the movie, first try to get info from db,
// if not exists, use polochon.Detailer and save informations in the database
// for future use
//
// If force is used, the detailer will be used even if the movie is found in
// database
func (m *Movie) GetDetails(env *web.Env, force bool) error {
	if len(m.Detailers) == 0 {
		m.Detailers = env.Config.MovieDetailers
	}

	log := env.Log.WithFields(logrus.Fields{
		"imdb_id":  m.ImdbID,
		"function": "movies.GetDetails",
	})

	// If the movie is not in db, we should add it, otherwise we should update
	// it
	var dbFunc func(db *sqlx.DB) error

	var err error
	err = m.Get(env)
	switch err {
	case nil:
		log.Debug("movie found in database")
		dbFunc = m.Update
		if !force {
			return nil
		}
	case ErrNotFound:
		dbFunc = m.Add
		log.Debug("movie not found in database")
	default:
		// Unexpected error
		return err
	}

	// so we got ErrNotFound so GetDetails from a detailer
	err = m.Movie.GetDetails(env.Log)
	if err != nil {
		return err
	}

	log.Debug("got details from detailers")

	err = dbFunc(env.Database)
	if err != nil {
		return err
	}

	log.Debug("movie added in database")

	// Download poster
	err = web.Download(m.Thumb, m.imgFile(env))
	if err != nil {
		return err
	}

	log.Debug("poster downloaded")

	// Set the poster url
	m.PosterURL = m.GetPosterURL(env)

	return nil
}

// Add a movie in the database
func (m *Movie) Add(db *sqlx.DB) error {
	var id string
	r, err := db.NamedQuery(addMovieQuery, m)
	if err != nil {
		return err
	}
	for r.Next() {
		r.Scan(&id)
	}
	m.ID = id

	return nil
}

// Update a movie in the database
func (m *Movie) Update(db *sqlx.DB) error {
	_, err := db.NamedQuery(updateMovieQuery, m)
	return err
}

// Delete movie from database
func (m *Movie) Delete(db *sqlx.DB) error {
	r, err := db.Exec(deleteMovieQuery, m.ID)
	if err != nil {
		return err
	}
	count, _ := r.RowsAffected()
	if count != 1 {
		return fmt.Errorf("Unexpected number of row deleted: %d", count)
	}
	return nil
}

// imgURL returns the default image url
func (m *Movie) imgURL(env *web.Env) string {
	return fmt.Sprintf("img/movies/%s.jpg", m.ImdbID)
}

// imgFile returns the image location on disk
func (m *Movie) imgFile(env *web.Env) string {
	return filepath.Join(env.Config.PublicDir, m.imgURL(env))
}

// GetPosterURL returns the image URL or the default image if the poster is not yet downloaded
func (m *Movie) GetPosterURL(env *web.Env) string {
	// Check if the movie image exists
	if _, err := os.Stat(m.imgFile(env)); os.IsNotExist(err) {
		// TODO image in the config ?
		return "img/noimage.png"
	}
	return m.imgURL(env)
}
